@@ -3,9 +3,11 @@ from typing import Optional
 from os.path import split, splitext
 
 from eis1600.markdown.UIDs import UIDs
-from eis1600.markdown.re_patterns import HEADER_END_SPLIT_PATTERN, MIU_LIGHT_OR_EIS1600_PATTERN, SPACES_PATTERN, \
-    NEWLINES_PATTERN, POETRY_PATTERN, SPACES_AFTER_NEWLINES_PATTERN, PAGE_TAG_ON_NEWLINE_PATTERN, UID_PATTERN, \
-    HEADING_OR_BIO_PATTERN, POETRY_TO_PARAGRAPH, BIO_CHR_TO_NEWLINE_PATTERN
+from eis1600.markdown.re_patterns import EMPTY_FIRST_PARAGRAPH_PATTERN, EMPTY_PARAGRAPH_PATTERN, HEADER_END_PATTERN, \
+    MIU_LIGHT_OR_EIS1600_PATTERN, ONLY_PAGE_TAG_PATTERN, PAGE_TAG_IN_BETWEEN_PATTERN, PAGE_TAG_PATTERN, \
+    PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN, SPACES_CROWD_PATTERN, NEWLINES_CROWD_PATTERN, \
+    POETRY_PATTERN, SPACES_AFTER_NEWLINES_PATTERN, PAGE_TAG_ON_NEWLINE_PATTERN, UID_PATTERN, HEADING_OR_BIO_PATTERN, \
+    BIO_CHR_TO_NEWLINE_PATTERN
 
 
 def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose: bool = False) -> None:
@@ -34,29 +36,35 @@ def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose
     with open(infile, 'r', encoding='utf8') as infile_h:
         text = infile_h.read()
 
-    header_and_text = HEADER_END_SPLIT_PATTERN.split(text)
+    header_and_text = HEADER_END_PATTERN.split(text)
     header = header_and_text[0] + header_and_text[1]
-    text = header_and_text[2]
+    text = header_and_text[2][1:]   # Ignore second new line after #META#Header#End#
+
+    if text[0:2] == '#\n':
+        # Some texts start with a plain #, remove these
+        text = text[2:]
 
     # fix
     text = text.replace('~\n', '\n')
     text = text.replace('\n~~', ' ')
+    text = text.replace(' \n', '\n')
 
     # spaces
     text = SPACES_AFTER_NEWLINES_PATTERN.sub('\n', text)
-    text = SPACES_PATTERN.sub(' ', text)
+    text = SPACES_CROWD_PATTERN.sub(' ', text)
 
     # fix poetry
     text = POETRY_PATTERN.sub(r'\1', text)
-    text = POETRY_TO_PARAGRAPH.sub(r'\1\n\n\2', text)
 
     # fix page tag on newlines
-    text = PAGE_TAG_ON_NEWLINE_PATTERN.sub(r' \1', text)
+    text = PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.sub(r'\1 \2 \3', text)
+    text = PAGE_TAG_ON_NEWLINE_PATTERN.sub(r' \1\n', text)
+    text = SPACES_CROWD_PATTERN.sub(' ', text)
 
+    # fix new lines
     text = text.replace('\n###', '\n\n###')
     text = text.replace('\n# ', '\n\n')
-
-    text = NEWLINES_PATTERN.sub('\n\n', text)
+    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
 
     text = text.split('\n\n')
 
@@ -68,9 +76,8 @@ def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose
             paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
         text_updated.append(paragraph)
 
-    text = '\n\n'.join(text_updated)
-
     # reassemble text
+    text = '\n\n'.join(text_updated)
     final = header + '\n\n' + text
 
     with open(outfile, 'w', encoding='utf8') as outfile_h:
@@ -78,7 +85,7 @@ def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose
 
 
 def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional[bool] = False) -> None:
-    """Insert UIDs into EIS1600TMP file and thereby convert it to EIS1600.
+    """Insert UIDs and EIS1600 tags into EIS1600TMP file and thereby convert it to EIS1600 format.
 
 
     :param str infile: Path of the file to convert.
@@ -102,10 +109,11 @@ def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional
     with open(infile, 'r', encoding='utf8') as infile_h:
         text = infile_h.read()
 
-    header_and_text = HEADER_END_SPLIT_PATTERN.split(text)
+    # disassemble text into paragraphs
+    header_and_text = HEADER_END_PATTERN.split(text)
     header = header_and_text[0] + header_and_text[1]
-    text = header_and_text[2]
-    text = NEWLINES_PATTERN.sub('\n\n', text)
+    text = header_and_text[2][1:]   # Ignore second new line after #META#Header#End#
+    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
     text = text.split('\n\n')
     text_updated = []
 
@@ -113,6 +121,9 @@ def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional
 
     text_iter = text.__iter__()
     paragraph = next(text_iter)
+    prev_p = ''
+
+    # Insert UIDs and EIS1600 tags into the text
     while paragraph is not None:
         next_p = next(text_iter, None)
 
@@ -121,22 +132,42 @@ def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional
             if paragraph.startswith('#'):
                 paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                 if next_p and not next_p.startswith('#'):
-                    heading_and_text = paragraph.split('\n', 1)
+                    # Insert a paragraph tag if there a multiple paragraphs in the current element
+                    heading_and_text = paragraph.splitlines()
                     if len(heading_and_text) > 1:
                         paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n' + \
                                     heading_and_text[1]
+
+                text_updated.append(paragraph)
             elif '%~%' in paragraph:
                 paragraph = f'_ء_={uids.get_uid()}= ::POETRY:: ~\n' + paragraph
+                text_updated.append(paragraph)
+            elif PAGE_TAG_PATTERN.fullmatch(paragraph):
+                page_tag = PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
+                if PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
+                    if text_updated:
+                        if text_updated[-1][-1] == ' ':
+                            text_updated[-1] += page_tag + ' ' + next_p
+                        else:
+                            text_updated[-1] += ' ' + page_tag + ' ' + next_p
+                        prev_p = paragraph
+                        paragraph = next_p
+                        next_p = next(text_iter, None)
+                elif text_updated:
+                    text_updated[-1] += ' ' + page_tag
+                # else:
+                    # Remove PageV00P000 at the beginning in an individual paragraph
+                    # text_updated.append(paragraph)
+                    # pass
             else:
                 paragraph = f'_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n' + paragraph
+                text_updated.append(paragraph)
 
-            text_updated.append(paragraph)
-
+        prev_p = paragraph
         paragraph = next_p
 
-    text = '\n\n'.join(text_updated)
-
     # reassemble text
+    text = '\n\n'.join(text_updated)
     final = header + '\n\n' + text
 
     with open(outfile, 'w', encoding='utf8') as outfile_h:
@@ -144,7 +175,7 @@ def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional
 
 
 def update_uids(infile: str, verbose: Optional[bool] = False) -> None:
-    """Updates a text with missing UIDs.
+    """Updates a text with missing UIDs and EIS1600 tags.
 
     :param str infile: Path of the file to update UIDs in.
     :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
@@ -161,14 +192,17 @@ def update_uids(infile: str, verbose: Optional[bool] = False) -> None:
     with open(infile, 'r', encoding='utf8') as infile_h:
         text = infile_h.read()
 
-    header_and_text = HEADER_END_SPLIT_PATTERN.split(text)
+    # disassemble text into paragraphs
+    header_and_text = HEADER_END_PATTERN.split(text)
     header = header_and_text[0] + header_and_text[1]
-    text = header_and_text[2]
+    text = header_and_text[2][1:]   # Ignore second new line after #META#Header#End#
+    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
     text = text.split('\n\n')
     text_updated = []
 
     used_ids = []
 
+    # Collect existing UIDs
     for idx, paragraph in enumerate(text):
         if UID_PATTERN.match(paragraph):
             uid = int(UID_PATTERN.match(paragraph).group('UID'))
@@ -187,15 +221,23 @@ def update_uids(infile: str, verbose: Optional[bool] = False) -> None:
 
     text_iter = text.__iter__()
     paragraph = next(text_iter)
+    prev_p = None
+
+    if EMPTY_FIRST_PARAGRAPH_PATTERN.fullmatch(paragraph):
+        # Some OpenITI texts start with a single # which causes a plain UID tag as first element - ignore those
+        paragraph = next(text_iter)
+
+    # Insert missing UIDs and EIS1600 tags into the text
     while paragraph is not None:
         next_p = next(text_iter, None)
 
         if paragraph:
-            # Only do this is paragraph is not empty
+            # Only do this if paragraph is not empty
             if HEADING_OR_BIO_PATTERN.match(paragraph):
                 paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                 if next_p and not MIU_LIGHT_OR_EIS1600_PATTERN.match(next_p):
-                    heading_and_text = paragraph.split('\n', 1)
+                    # Insert a paragraph tag if there a multiple paragraphs in the current element
+                    heading_and_text = paragraph.splitlines()
                     if len(heading_and_text) > 1:
                         paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n' + \
                                     heading_and_text[1]
@@ -203,12 +245,34 @@ def update_uids(infile: str, verbose: Optional[bool] = False) -> None:
                 section_header = '' if paragraph.startswith('::') else '::UNDEFINED:: ~\n'
                 paragraph = f'_ء_={uids.get_uid()}= {section_header}' + paragraph
 
-            text_updated.append(paragraph)
+            if ONLY_PAGE_TAG_PATTERN.fullmatch(paragraph):
+                # Add page tags to previous paragraph if there is no other information contained in the current
+                # paragraph
+                page_tag = ONLY_PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
+                if PAGE_TAG_IN_BETWEEN_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
+                    if text_updated:
+                        next_p_text = next_p.split('::UNDEFINED:: ~\n')[1]
+                        if text_updated[-1][-1] == ' ':
+                            text_updated[-1] += page_tag + ' ' + next_p_text
+                        else:
+                            text_updated[-1] += ' ' + page_tag + ' ' + next_p_text
+                        prev_p = paragraph
+                        paragraph = next_p
+                        next_p = next(text_iter, None)
+                elif text_updated:
+                    text_updated[-1] += ' ' + page_tag
+                # else:
+                    # Remove PageV00P000 at the beginning in an individual paragraph
+                    # text_updated.append(paragraph)
+                    # pass
+            elif not EMPTY_PARAGRAPH_PATTERN.fullmatch(paragraph):
+                # Do not add empty paragraphs to the updated text
+                text_updated.append(paragraph)
+        prev_p = paragraph
         paragraph = next_p
 
-    text = '\n\n'.join(text_updated)
-
     # reassemble text
+    text = '\n\n'.join(text_updated)
     final = header + '\n\n' + text
 
     with open(outfile, 'w', encoding='utf8') as outfile_h:
